#pragma once

#include <cglib/core/camera.h>
#include <cglib/gl/shader.h>
#include <cglib/gl/scene_graph.h>

#include <glm/glm.hpp>

// DO NOT PUT ANY OPENGL CALLS INTO CONSTRUCTOR!!!
// all intialization must be done in initialize!

class DeviceRenderingParameters;
class GLObjModel;
struct FBO;
struct AABB;

/*
 * Call the drawing functor, but first set up the OpenGL context so that
 * all geometry is drawn as a wireframe overlay.
 */
template <typename DrawFunc>
void wireframe_overlay(DrawFunc&& draw)
{
	// Store a copy of the current OpenGL state for all attributes that
	// we are going to modify.
	glPushAttrib(GL_DEPTH_BUFFER_BIT | GL_POLYGON_BIT);

	// Disable depth testing and writing (this is an overlay), and draw as lines.
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	// Slightly thicker lines for
	//glLineWidth(2.f);

	// Draw using the provided functor.
	draw();

	// Restore OpenGL state.
	glPopAttrib();
}

class Renderer
{
public:
	ShaderManager shader_manager;
	std::shared_ptr<Camera> camera;
	Renderer(const MapNameProgramDefinition &program_definitions)
		: shader_manager(program_definitions)
	{
	}
	virtual ~Renderer()       {}
	virtual void draw()       = 0;
	virtual void destroy()    {}
	virtual void activate()   {}
	virtual void deactivate() {}
	virtual int  initialize(const DeviceRenderingParameters &parameters);
	virtual int  update_parameters(const DeviceRenderingParameters &parameters);
	virtual void set_active_camera();
	virtual void update_objects(float time_step);

	glm::mat4 projection;
};

class CubeRenderer : public Renderer
{
public:
	CubeRenderer()
		: Renderer(program_definitions)
	{
	}
	virtual void draw() override;
	virtual void destroy() override;
	virtual void activate() override;
	virtual void deactivate() override;
	virtual int  initialize(const DeviceRenderingParameters &parameters) override;

	unsigned int buffer_cube_vert, buffer_cube_idx, vao_cube;

	static const MapNameProgramDefinition program_definitions;
};

class MonkeyRenderer : public Renderer
{
public:
	MonkeyRenderer()
		: Renderer(program_definitions)
	{
	}
	virtual void draw() override;
	virtual void activate() override;
	virtual void deactivate() override;
	virtual int  initialize(const DeviceRenderingParameters &parameters) override;
	virtual void destroy() override;

	std::shared_ptr<GLObjModel> monkey, fullscreen_tri;
	unsigned envTex, envTexGlossy, envTexDiffuse;
	const static MapNameProgramDefinition program_definitions;
};

class HDRRenderer
{
public:
	HDRRenderer() :
		shader_manager(program_definitions)
	{}

	int  initialize(const DeviceRenderingParameters &parameters);
	int  update_parameters(const DeviceRenderingParameters &parameters);
	void destroy();
	void tonemap();
	void begin_hdr_rendering();
	void end_hdr_rendering();

	ShaderManager shader_manager;
	std::shared_ptr<GLObjModel> fullscreen_tri;
	std::shared_ptr<FBO> fbo;
	const static MapNameProgramDefinition program_definitions;
};

class SponzaRenderer : public Renderer
{
public:
	SponzaRenderer() :
		Renderer(program_definitions)
	{}

	virtual int  initialize(const DeviceRenderingParameters &parameters) override;
	virtual int  update_parameters(const DeviceRenderingParameters &parameters) override;
	virtual void draw() override;
	virtual void activate() override;
	virtual void deactivate() override;
	virtual void destroy() override;

	virtual void update_objects(float time_step) override;

	std::shared_ptr<GLObjModel> sponza;
	const static MapNameProgramDefinition program_definitions;

	bool enable_animation = false;
	float time = 0.0f;

	HDRRenderer hdrRenderer;

	enum { LIGHT_COUNT = 9 };
	glm::vec3 light_pos[LIGHT_COUNT];
	glm::vec3 light_intensity[LIGHT_COUNT];
};

class SphereFlakeRenderer : public Renderer
{
private:
	std::shared_ptr<SceneGraphNode> sceneGraphRoot;
	std::shared_ptr<GLObjModel> sphereModel;
	bool enable_animation = false;
	float animationTime = 0.0f;
	float animationSpeedFactor = 1.0f;
	int sphere_flake_recursion_depth = 0;
	float sphere_flake_size_factor = 0.5f;

	std::vector<TransformedModel> transformedRenderingObjects;

	const static MapNameProgramDefinition program_definitions;
public:
	SphereFlakeRenderer() :
		Renderer(program_definitions)
	{}

	virtual void activate() override;
	virtual void deactivate() override;
	virtual void destroy() override;
	virtual int initialize(const DeviceRenderingParameters &parameters) override;
	virtual void draw() override;
	virtual void update_objects(float time_step) override;
	virtual int  update_parameters(const DeviceRenderingParameters &parameters) override;
};

