#include <cglib/core/glheaders.h>
#include <cglib/gl/glmodel.h>
#include <cglib/core/obj_mesh.h>
#include <cglib/core/image.h>

#include <iostream>
#include <map>
#include <unordered_map>

struct GLTextureList
{
	std::map<std::string, unsigned int> textureIDs;

	GLTextureList() :
		textureIDs() {}

	unsigned int getGLTexture(const std::string& filename, float gamma = 2.2f);
};

unsigned int GLTextureList::getGLTexture(const std::string& filename, float gamma) {
	const auto& i = textureIDs.find(filename);

	// already loaded?
	if (i != textureIDs.end())
		return i->second;

	// load texture
	//std::cout << "loading texture: " << filename << std::endl;

	GLint internalFormat = GL_RGBA;
	if (gamma == 2.2f) {
		internalFormat = GL_SRGB_ALPHA;
		gamma = 1.0f;
	}

	Image img;
	img.load(filename, gamma);

	unsigned int glTex;
	glGenTextures(1, &glTex);

	glBindTexture(GL_TEXTURE_2D, glTex);

	glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, img.getWidth(), img.getHeight(), 0, GL_RGBA, GL_FLOAT, img.getPixels());

	glGenerateMipmap(GL_TEXTURE_2D);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	glBindTexture(GL_TEXTURE_2D, 0);

	// insert into map
	textureIDs[filename] = glTex;

	return glTex;
}

GLObjModel::
GLObjModel(const std::string &path)
{
	OBJFile obj_file(false);
	bool result = obj_file.loadFile(path);
	if(!result) {
		std::cerr << "could not load model \"" << path << "\"!" << std::endl;
		return;
	}
	glGenVertexArrays(1, &vao);
	glGenBuffers(1, &buf_vertex);
	glGenBuffers(1, &buf_normal);
	glGenBuffers(1, &buf_texcoord);
	glGenBuffers(1, &buf_index);

	std::vector<glm::vec3> vertices, normals;
	std::vector<glm::vec2> tex_coords;
	std::vector<uint32_t> indices;

	auto hash_func = [](const glm::uvec3 &v) -> size_t {
			std::hash<uint32_t> hasher;
			size_t seed = hasher(v.x);
			seed ^= hasher(v.y) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
			seed ^= hasher(v.z) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
			return seed;
	};

	std::unordered_map<glm::uvec3, uint32_t, decltype(hash_func)&> vertex_index_map(
			100000, hash_func
			);

	GLTextureList textures;

	for(unsigned i = 0; i < obj_file.getModelCount(); i++) {
		const auto &model = obj_file.getModel(i);
		vertex_index_map.clear();

		const auto &modelSurfaces = model->getSurfaces();
		for(const auto &surf: modelSurfaces) {

			GLObjModelSurface surfaceInfo;
			// save indices offset
			surfaceInfo.startIndex = indices.size();

			// get texture
			const auto iter = surf->material->additionalInfo.find("map_Kd");

			if (iter != surf->material->additionalInfo.end())
			{
				surfaceInfo.gl_texBuffer = textures.getGLTexture(iter->second);
			}
			else {
				surfaceInfo.gl_texBuffer = textures.getGLTexture("assets/empty.png");
			}

			for(unsigned j = 0; j < surf->vertexIndices.size(); j++) {
				for(int k = 0; k < 3; k++) {
					glm::uvec3 idx(~0);
					idx[0] = surf->vertexIndices[j][k];
					if(surf->normalIndices.size() > 0) {
						idx[1] = surf->normalIndices.at(j)[k];
					}
					if(surf->texcoordIndices.size() > 0) {
						idx[2] = surf->texcoordIndices.at(j)[k];
					}

					uint32_t vertex_idx;
					auto it = vertex_index_map.find(idx);
					if(it != vertex_index_map.end()) {
						vertex_idx = it->second;
					}
					else {
						vertices.push_back(model->getVertices().at(idx[0]));
						if(surf->normalIndices.size() > 0) {
							normals.push_back(model->getNormals().at(idx[1]));
						}
						if(surf->texcoordIndices.size() > 0) {
							tex_coords.push_back(model->getTexcoords().at(idx[2]));
						}
						vertex_idx = vertices.size() - 1;
						vertex_index_map[idx] = vertex_idx;
					}
					indices.push_back(vertex_idx);
				}
			}

			// save end index and store to surface list
			surfaceInfo.endIndex = indices.size();
			modelSurfaceRange.push_back(surfaceInfo);
		}
	}

	// print surface info:
	for (unsigned int i = 0 ; i < modelSurfaceRange.size(); ++i)
	{
		//std::cout << "model " << i << " -> indices [" << modelSurfaceRange[i].startIndex << ", " << modelSurfaceRange[i].endIndex << "]" << std::endl;
	}

	num_indices = indices.size();

	glBindVertexArray(vao);
	glBindBuffer(GL_ARRAY_BUFFER, buf_vertex);
	glBufferData(GL_ARRAY_BUFFER, sizeof(glm::vec3) * vertices.size(), vertices.data(), GL_STATIC_DRAW);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

	if(normals.size() > 0) {
		glBindBuffer(GL_ARRAY_BUFFER, buf_normal);
		glBufferData(GL_ARRAY_BUFFER, sizeof(glm::vec3) * normals.size(), normals.data(), GL_STATIC_DRAW);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	}

	if(tex_coords.size() > 0) {
		glBindBuffer(GL_ARRAY_BUFFER, buf_texcoord);
		glBufferData(GL_ARRAY_BUFFER, sizeof(glm::vec2) * tex_coords.size(), tex_coords.data(), GL_STATIC_DRAW);
		glEnableVertexAttribArray(2);
		glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 0, 0);
	}

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buf_index);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, num_indices * sizeof(uint32_t), indices.data(), GL_STATIC_DRAW);

	glBindVertexArray(0);
}

GLObjModel::
~GLObjModel()
{
	glDeleteVertexArrays(1, &vao);
	glDeleteBuffers(1, &buf_vertex);
	glDeleteBuffers(1, &buf_normal);
	glDeleteBuffers(1, &buf_texcoord);
	glDeleteBuffers(1, &buf_index);
}

void GLObjModel::
draw()
{
	glBindVertexArray(vao);

	// draw each surface with texture
	for (const GLObjModelSurface& faceData : modelSurfaceRange)
	{
		glBindTexture(GL_TEXTURE_2D, faceData.gl_texBuffer);

		unsigned int numIndices = faceData.endIndex - faceData.startIndex;
		glDrawElements(GL_TRIANGLES, numIndices, GL_UNSIGNED_INT, (char*)0 + sizeof(unsigned int) * faceData.startIndex);
	}

	glBindVertexArray(0);
}

int GLObjModel::
index_count() const
{
	int vc = 0;
	for(auto &i: modelSurfaceRange) {
		vc += (i.endIndex - i.startIndex);
	}
	return vc;
}
